import numpy as np
from anesthetic import MCMCSamples
from margarine.maf import MAF
from margarine.marginal_stats import calculate
import pytest
from margarine.kde import KDE
from margarine.clustered import clusterMAF
from numpy.testing import assert_equal, assert_allclose
from scipy.stats import ks_2samp

def likelihood(parameters):
        y = np.array([0, 6])
        loglikelihood = (-0.5*np.log(2*np.pi*(1**2))-0.5 \
            *((y - parameters) / 1)**2).sum(axis=1)
        return loglikelihood

def D_KL(logL, weights):
    return -np.average(logL, weights=weights)

def d_g(logL, weights):
    return 2*np.cov(logL, aweights=weights)

ndims = 2
nsamples = 2500
x = np.random.normal(0, 1, nsamples)
y = np.random.normal(6, 1, nsamples)

theta = np.vstack([x, y]).T
weights = np.ones(len(theta))

logL = likelihood(theta)

mcmc_samples = MCMCSamples(data=theta, logL=logL)
samples_kl = D_KL(logL, weights)
samples_d = d_g(logL, weights)
names = [i for i in range(theta.shape[-1])]

def test_maf():

    def check_stats(i):
        if i ==0:
            value = samples_kl
        else:
            value = samples_d
        assert_allclose(stats['Value'][i], value, rtol=1, atol=1)

    bij = MAF(theta, weights=weights)
    bij.train(10000, early_stop=True)

    stats = calculate(bij).statistics()
    [check_stats(i) for i in range(2)]

    equal_weight_theta = mcmc_samples.compress(50)[names].values
    x = bij.sample(len(equal_weight_theta))

    res = [ks_2samp(equal_weight_theta[:, i], x[:, i]) 
           for i in range(equal_weight_theta.shape[-1])]
    p_values = [res[i].pvalue for i in range(len(res))]
    for i in range(len(p_values)):
        assert(round(p_values[i], 2) >0.05)

    estL = bij.log_like(equal_weight_theta.astype(np.float32), 0.0)
    check_like = 0
    for i in range(len(logL)):
        if np.isclose(np.exp(logL[i]), np.exp(estL[i]), rtol=1, atol=1):
            check_like += 1
    assert((len(logL)-check_like)/len(logL) <= 0.1)

def test_maf_kwargs():

    with pytest.raises(TypeError):
        bij = MAF(theta, weights=weights)
        bij.sample(4.5)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights, number_networks=4.4)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights, learning_rate='foobar')
    with pytest.raises(TypeError):
        MAF(theta, weights=weights, hidden_layers='foobar')
    with pytest.raises(TypeError):
        MAF(theta, weights=weights, hidden_layers=[4.5, 50])
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=4.5)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, early_stop='foo')
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, clustering=5)
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, cluster_numeber='foo')
    with pytest.raises(TypeError):
        MAF(theta, weights=weights)
        bij.train(epochs=100, cluster_labels=5)

def test_maf_save_load():

    bij = MAF(theta, weights=weights)
    bij.train(100)
    file = 'saved_maf.pkl'
    bij.save(file)
    loaded_bijector = MAF.load(file)
    for i in range(len(bij.mades)):
        assert_equal(bij.mades[i].get_weights(),
            loaded_bijector.mades[i].get_weights())

def test_kde():

    def check_stats(i):
        if i ==0:
            value = samples_kl
        else:
            value = samples_d
        assert_allclose(stats['Value'][i], value, rtol=1, atol=1)

    kde = KDE(theta, weights=weights)
    kde.generate_kde()

    stats = calculate(kde).statistics()
    [check_stats(i) for i in range(2)]

    equal_weight_theta = mcmc_samples.compress(50)[names].values
    x = kde.sample(len(equal_weight_theta))

    res = [ks_2samp(equal_weight_theta[:, i], x[:, i]) 
           for i in range(equal_weight_theta.shape[-1])]
    p_values = [res[i].pvalue for i in range(len(res))]
    for i in range(len(p_values)):
        assert(round(p_values[i], 2) >0.05)

    estL = kde.log_like(equal_weight_theta, 0.0)
    check_like = 0
    for i in range(len(logL)):
        if np.isclose(np.exp(logL[i]), np.exp(estL[i]), rtol=1, atol=1):
            check_like += 1
    assert((len(logL)-check_like)/len(logL) <= 0.1)

def test_kde_save_load():

    kde = KDE(theta, weights=weights)
    kde.generate_kde()
    file = 'saved_maf.pkl'
    kde.save(file)
    loaded_kde = KDE.load(file)
    assert_equal(kde.kde.covariance, loaded_kde.kde.covariance)

def test_anesthetic():

    kde = KDE(mcmc_samples, parameters=names)
    maf = MAF(mcmc_samples, parameters=names)
    cmaf = clusterMAF(mcmc_samples)

    assert_equal(kde.parameters, names)
    assert_equal(maf.parameters, names)

    # not providing parametes here but deriving them from the 
    # anesthetic object columns
    assert(np.all(cmaf.parameters == np.array(names)))



import numpy as np
from tensorflow_probability import distributions as tfd
import pandas as pd
from margarine.maf import MAF


class calculate(object):

    r"""

    This class, once initalised with a trained MAF or KDE and samples,
    can be used to calculate marginal KL divergences and
    bayesian dimensionalities.

    **Paramesters:**

        de: **instance of MAF class or KDE class**
            | This should be a loaded and trained instance of a MAF, clusterMAF
                or KDE.Bijectors can be loaded like so

                .. code:: python

                    from margarine.maf import MAF
                    from margarine.kde import KDE
                    from margarine.clustered import clusterMAF

                    file = '/trained_maf.pkl'
                    maf = MAF.load(file)

                    file = '/trained_kde.pkl'
                    kde = KDE.load(file)

                    file = '/trained_clustered_maf.pkl'
                    clustered_maf = clusterMAF.load(file)

        samples: **numpy array**
            | This should be the output of the bijector when called to generate
                a set of samples from the replicated probability
                distribution. e.g. after loading a trained MAF we would pass

                .. code:: python

                    u = np.random.uniform(0, 1, size=(10000, 5))
                    prior_limits = np.array([[0]*5, [1]*5])
                    samples = maf(u, prior_limits)

    **Kwargs:**

        prior_de: **instance of MAF class, clusterMAF class or KDE class**
            | This should be a loaded and trained instance of a MAF, clusterMAF
                or KDE for the prior.
                If not provided, a uniform prior will be used.

    """

    def __init__(self, de, **kwargs):

        self.de = de

        self.theta = self.de.theta
        self.theta_weights = self.de.sample_weights
        self.theta_min = self.de.theta_min
        self.theta_max = self.de.theta_max

        if isinstance(self.de, MAF):
            self.theta_weights = self.theta_weights.numpy()
            self.theta_min = self.theta_min.numpy()
            self.theta_max = self.theta_max.numpy()

        self.samples = self.de.sample(len(self.theta))

        self.prior_de = kwargs.pop('prior_de', None)

    def statistics(self):

        """
        Calculate marginal bayesian KL divergence and dimensionality with
        approximate errors.
        """

        def mask_arr(arr):
            return arr[np.isfinite(arr)], np.isfinite(arr)

        logprob = self.de.log_prob(self.samples)
        theta_logprob = self.de.log_prob(self.theta)

        if isinstance(self.de, MAF):
            logprob = logprob.numpy()
            theta_logprob = theta_logprob.numpy()

        args = np.argsort(theta_logprob)
        self.theta_weights = self.theta_weights[args]
        theta_logprob = theta_logprob[args]

        deargs = np.argsort(logprob)
        logprob = logprob[deargs]
        wde = [np.sum(self.theta_weights)/len(logprob)]*len(logprob)
        logprob = np.interp(
            np.cumsum(self.theta_weights), np.cumsum(wde), logprob)

        mid_point = np.log((np.exp(logprob) + np.exp(theta_logprob))/2)

        if self.prior_de is None:
            self.base = tfd.Blockwise(
                [tfd.Uniform(self.theta_min[i], self.theta_max[i])
                 for i in range(self.samples.shape[-1])])
            prior = self.base.sample(len(self.theta)).numpy()

            theta_base_logprob = self.base.log_prob(self.theta).numpy()
            base_logprob = self.base.log_prob(prior).numpy()

            prior_wde = [np.sum(self.theta_weights)/len(base_logprob)] * \
                len(base_logprob)

            base_logprob = base_logprob[deargs]
            base_logprob = np.interp(
                np.cumsum(self.theta_weights), np.cumsum(prior_wde),
                base_logprob)

            theta_base_logprob = theta_base_logprob[args]
        elif self.prior_de is not None:
            self.base = self.prior_de.copy()
            de_prior_samples = self.base.sample(len(self.theta))
            theta_base_logprob = self.base.log_prob(de_prior_samples)
            base_logprob = self.base.log_prob(self.theta)

            if isinstance(self.prior_de, MAF):
                theta_base_logprob = theta_base_logprob.numpy()
                base_logprob = base_logprob.numpy()

            base_logprob = base_logprob[deargs]
            base_logprob = np.interp(
                np.cumsum(self.theta_weights), np.cumsum(wde), base_logprob)

            base_args = np.argsort(theta_base_logprob)
            theta_base_logprob = theta_base_logprob[base_args]
            prior_weights = np.cumsum(self.prior_de.sample_weights[base_args])

            prior_weights = (np.cumsum(self.theta_weights).max() -
                             np.cumsum(self.theta_weights).min()) * \
                            (prior_weights - prior_weights.min()) / \
                            (prior_weights.max() - prior_weights.min()) + \
                np.cumsum(self.theta_weights).min()

            theta_base_logprob = np.interp(np.cumsum(self.theta_weights),
                                           prior_weights, theta_base_logprob)

        midbasepoint = np.log((np.exp(base_logprob) +
                               np.exp(theta_base_logprob))/2)

        mid_logL, mask = mask_arr(mid_point - midbasepoint)
        midkl = np.average(mid_logL, weights=self.theta_weights[mask])
        midbd = 2*np.cov(mid_logL, aweights=self.theta_weights[mask])
        de_logL, mask = mask_arr(logprob - base_logprob)
        dekl = np.average(de_logL, weights=self.theta_weights[mask])
        debd = 2*np.cov(de_logL, aweights=self.theta_weights[mask])
        theta_logL, mask = mask_arr(theta_logprob - theta_base_logprob)
        thetakl = np.average(theta_logL, weights=self.theta_weights[mask])
        thetabd = 2*np.cov(theta_logL, aweights=self.theta_weights[mask])

        kl_array = np.sort([midkl, dekl, thetakl])
        bd_array = np.sort([midbd, debd, thetabd])

        results_dict = {'Statistic': ['KL Divergence', 'BMD'],
                        'Value': [kl_array[1], bd_array[1]],
                        'Lower Bound': [kl_array[0], bd_array[0]],
                        'Upper Bound': [kl_array[2], bd_array[2]]}

        results = pd.DataFrame(results_dict)
        results.set_index('Statistic', inplace=True)

        return results



from scipy.special import logsumexp
from sklearn.cluster import KMeans
from tensorflow import keras
from margarine.maf import MAF
import anesthetic
import numpy as np
import tensorflow as tf
import warnings
import pickle


class clusterMAF():

    r"""

    This class is used to train, load and call a piecewise normalising flow
    built from a set of masked autoregressive flows. The class
    is essentially a wrapper around the MAF class with some additional
    clustering functionality. This class has all the same
    functionality as the MAF class and can be used in the same way.

    **Parameters:**

        theta: **numpy array or anesthetic.samples**
            | The samples from the probability distribution that we require the
                MAF to learn. This can either be a numpy array or an anesthetic
                NestedSamples or MCMCSamples object.

    **kwargs:**

        weights: **numpy array / default=np.ones(len(theta))**
            | The weights associated with the samples above. If an anesthetic
                NestedSamples or MCMCSamples object is passed the code
                draws the weights from this.

        number_networks: **int / default = 6**
            | The bijector is built by chaining a series of
                autoregressive neural
                networks together and this parameter is used to determine
                how many networks there are in the chain.

        learning_rate: **float / default = 1e-3**
            | The learning rate determines the 'step size' of the optimization
                algorithm used to train the MAF. Its value can effect the
                quality of emulation.

        hidden_layers: **list / default = [50, 50]**
            | The number of layers and number of nodes in each hidden layer for
                each neural network. The default is two hidden layers with
                50 nodes each and each network in the chain has the same hidden
                layer structure.

        cluster_labels: **list / default = None**
            | If clustering has been performed externally to margarine you can
                provide a list of labels for the samples theta. The labels
                should be integers from 0 to k corresponding to the cluster
                that each sample is in.

        cluster_number: **int / default = None**
            | If clustering has been performed externally to margarine you
                need to provide the number of clusters, k, alongside the
                cluster labels.

        parameters: **list of strings**
            | A list of the relevant parameters to train on. Only needed
                if theta is an anestehetic samples object. If not provided,
                all parameters will be used.

    """

    def __init__(self, theta, **kwargs):

        self.number_networks = kwargs.pop('number_networks', 6)
        self.learning_rate = kwargs.pop('learning_rate', 1e-3)
        self.hidden_layers = kwargs.pop('hidden_layers', [50, 50])
        self.cluster_labels = kwargs.pop('cluster_labels', None)
        self.cluster_number = kwargs.pop('cluster_number', None)
        self.parameters = kwargs.pop('parameters', None)

        if isinstance(theta, 
                      (anesthetic.samples.NestedSamples, 
                       anesthetic.samples.MCMCSamples)):
            weights = theta.get_weights()
            if self.parameters:
                theta = theta[self.parameters].values
            else:
                if isinstance(theta, anesthetic.samples.NestedSamples):
                    self.parameters = theta.columns[:-3].values
                    theta = theta[theta.columns[:-3]].values
                else:
                    self.parameters = theta.columns[:-1].values
                    theta = theta[theta.columns[:-1]].values
        else:
            weights = kwargs.pop('weights', np.ones(len(theta)))

        self.theta = theta
        self.sample_weights = weights

        # needed for marginal stats
        self.n = np.sum(self.sample_weights)**2 / \
            np.sum(self.sample_weights**2)
        theta_max = np.max(self.theta, axis=0)
        theta_min = np.min(self.theta, axis=0)
        a = ((self.n-2)*theta_max-theta_min)/(self.n-3)
        b = ((self.n-2)*theta_min-theta_max)/(self.n-3)
        self.theta_min = b
        self.theta_max = a

        if type(self.number_networks) is not int:
            raise TypeError("'number_networks' must be an integer.")
        if not isinstance(self.learning_rate,
                          (int, float,
                           keras.optimizers.schedules.LearningRateSchedule)):
            raise TypeError("'learning_rate', " +
                            "must be an integer, float or keras scheduler.")
        if type(self.hidden_layers) is not list:
            raise TypeError("'hidden_layers' must be a list of integers.")
        else:
            for i in range(len(self.hidden_layers)):
                if type(self.hidden_layers[i]) is not int:
                    raise TypeError(
                        "One or more valus in 'hidden_layers'" +
                        "is not an integer.")

        mask = np.isfinite(theta).all(axis=-1)
        self.theta = self.theta[mask]
        self.sample_weights = self.sample_weights[mask]

        if self.cluster_number is not None:
            if self.cluster_labels is None:
                raise ValueError("'cluster_labels' should be provided if " +
                                 "'cluster_number' is specified.")
        else:
            if self.cluster_labels is not None:
                raise ValueError("'cluster_number' should be provided if " +
                                 "'cluster_labels' is specified.")

        if self.cluster_number is not None:
            if self.cluster_number % 1 != 0:
                raise TypeError("'cluster_number' must be a whole number.")
        if self.cluster_labels is not None:
            if not isinstance(self.cluster_labels, (np.ndarray, list)):
                raise TypeError("'cluster_labels' must be an array " +
                                "or a list.")

        if self.cluster_number is None:
            # this code currently performs the clustering with
            # kmeans but we could have anything here.
            from sklearn.metrics import silhouette_score
            ks = np.arange(2, 21)
            losses = []
            for k in ks:
                kmeans = KMeans(k, random_state=0)
                labels = kmeans.fit(self.theta).predict(self.theta)
                losses.append(-silhouette_score(self.theta, labels))
            losses = np.array(losses)
            minimum_index = np.argmin(losses)
            self.cluster_number = ks[minimum_index]

            kmeans = KMeans(self.cluster_number, random_state=0)
            self.cluster_labels = kmeans.fit(self.theta).predict(self.theta)

        if self.cluster_number == 20:
            warnings.warn("The number of clusters is 20. " +
                          "This is the maximum " +
                          "number of clusters that can be used. If you " +
                          "require more clusters, please specify the " +
                          "'cluster_number' kwarg. " +
                          "margarine will continue " +
                          "with 20 clusters.")

        if np.array(list(self.cluster_labels)).dtype == 'float':
            # convert cluster labels to integers
            self.cluster_labels = self.cluster_labels.astype(int)
        # count the number of times a cluster label appears in cluster_labels
        self.cluster_count = np.bincount(self.cluster_labels)
        # While loop to make sure clusters are not too small
        while self.cluster_count.min() < 100:
            warnings.warn("One or more clusters are too small " +
                          "(n_cluster < 100). " +
                          "Reducing the number of clusters by 1.")
            minimum_index -= 1
            self.cluster_number = ks[minimum_index]
            kmeans = KMeans(self.cluster_number, random_state=0)
            self.cluster_labels = kmeans.fit(self.theta).predict(self.theta)
            self.cluster_count = np.bincount(self.cluster_labels)
            if self.cluster_number == 2:
                # break if two clusters
                warnings.warn("The number of clusters is 2. This is the " +
                              "minimum number of clusters that can be used. " +
                              "Some clusters may be too small and the " +
                              "train/test split may fail." +
                              "Try running without clusting. ")
                break

        split_theta = []
        split_sample_weights = []
        for i in range(self.cluster_number):
            split_theta.append(self.theta[self.cluster_labels == i])
            split_sample_weights.append(
                self.sample_weights[self.cluster_labels == i])

        self.split_theta = split_theta
        self.split_sample_weights = split_sample_weights

        self.flow = []
        for i in range(len(split_theta)):
            self.flow.append(MAF(split_theta[i], 
                                 weights=split_sample_weights[i],
                                 number_networks=self.number_networks,
                                 learning_rate=self.learning_rate,
                                 hidden_layers=self.hidden_layers))

    def train(self, epochs=100, early_stop=False, loss_type='sum'):

        r"""
        This function is called to train the clusterMAF once it has been
        initialised. It calls the `train()` function for each flow.

        .. code:: python

            from margarine.clustered import clusterMAF

            bij = clusterMAF(theta, weights=weights)
            bij.train()

        **Kwargs:**

            epochs: **int / default = 100**
                | The number of iterations to train the neural networks for.

            early_stop: **boolean / default = False**
                | Determines whether or not to implement an early stopping
                    algorithm or
                    train for the set number of epochs. If set to True then the
                    algorithm will stop training when test loss has not
                    improved for 2% of the requested epochs. At this point
                    margarine will roll back to the best model and return this
                    to the user.

            loss_type: **string / default = 'sum'**
                | Determines whether to use the sum or mean of the weighted
                    log probabilities to calculate the loss function.

        """

        for i in range(len(self.flow)):
            self.flow[i].train(epochs=epochs, early_stop=early_stop,
                               loss_type=loss_type)

    def log_prob(self, params):

        """
        Function to caluclate the log-probability for a given clusterMAF and
        set of parameters.

        While each density estimator has its own built in log probability
        function, a correction has to be applied for the transformation of
        variables that is used to improve accuracy when learning and we
        have to sum probabilities over the series of flows. The
        correction and the sum are implemented here.

        **Parameters:**

            params: **numpy array**
                | The set of samples for which to calculate the log
                    probability.

        """

        # currently working with numpy not tensorflow
        # nan repalacement with 0 difficult in tensorflow
        logprob = []
        for flow in self.flow:
            probs = flow.log_prob(params).numpy()
            for j in range(len(probs)):
                if np.isnan(probs[j]):
                    probs[j] = np.log(1e-300)
            logprob.append(probs)
        logprob = np.array(logprob)
        print(logprob.shape)
        logprob = logsumexp(logprob, axis=0)

        return logprob

    def log_like(self, params, logevidence, prior_de=None):

        r"""
        This function should return the log-likelihood for a given set of
        parameters.

        It requires the logevidence from the original nested sampling run
        in order to do this and in the case that the prior is non-uniform
        a trained prior density estimator should be provided.

        **Parameters:**

            params: **numpy array**
                | The set of samples for which to calculate the log
                    probability.

            logevidence: **float**
                | Should be the log-evidence from the full nested sampling
                    run with nuisance parameters.

            prior_de: **margarine.maf.MAF / default=None**
                | If the prior is non-uniform then a trained prior density
                    estimator should be provided. Otherwise the prior
                    is assumed to be uniform and the prior probability
                    is calculated analytically from the minimum and maximum
                    values of the parameters.

        """

        if prior_de is None:
            warnings.warn('Assuming prior is uniform!')

            n = (np.sum(self.sample_weights)**2) / \
                (np.sum(self.sample_weights**2))

            theta_max = np.max(self.theta, axis=0)
            theta_min = np.min(self.theta, axis=0)
            a = ((n-2)*theta_max-theta_min)/(n-3)
            b = ((n-2)*theta_min-theta_max)/(n-3)
            prior_logprob = np.log(
                np.prod([1/(a - b)
                         for i in range(len(b))]))
        else:
            prior_logprob = prior_de.log_prob(params).numpy()

        posterior_logprob = self.log_prob(params)

        loglike = posterior_logprob + prior_logprob - logevidence

        return loglike

    @tf.function(jit_compile=True)
    def __call__(self, u):

        r"""

        This function is used when calling the clusterMAF class to transform
        samples from the unit hypercube to samples on the clusterMAF.

        **Parameters:**

            u: **numpy array**
                | Samples on the uniform hypercube.

        """

        len_thetas = [len(self.split_theta[i])
                      for i in range(len(self.split_theta))]
        probabilities = [len_thetas[i]/np.sum(len_thetas)
                         for i in range(len(self.split_theta))]
        options = np.arange(0, self.cluster_number)
        choice = np.random.choice(options,
                                  p=probabilities, size=len(u))

        totals = [len(choice[choice == options[i]])
                  for i in range(len(options))]
        totals = np.hstack([0, np.cumsum(totals)])

        values = []
        for i in range(len(options)):
            x = self.flow[i](u[totals[i]:totals[i+1]])
            values.append(x)

        x = tf.concat(values, axis=0)
        return x

    def sample(self, length=1000):

        r"""

        This function is used to generate samples on the clusterMAF via the
        clusterMAF __call__ function.

        **Kwargs:**

            length: **int / default=1000**
                | This should be an integer and is used to determine how many
                    samples are generated when calling the clusterMAF.

        """

        u = np.random.uniform(size=(length, self.theta.shape[-1]))
        return self(u)

    def save(self, filename):

        r"""

        This function can be used to save an instance of
        a trained clusterMAF as
        a pickled class so that it can be loaded and used in differnt scripts.

        **Parameters:**

            filename: **string**
                | Path in which to save the pickled MAF.

        """

        nn_weights = {}
        for i in range(self.cluster_number):
            made = self.flow[i].mades
            w = [m.get_weights() for m in made]
            nn_weights[i] = w

        with open(filename, 'wb') as f:
            pickle.dump([self.theta,
                        nn_weights,
                        self.sample_weights,
                        self.number_networks,
                        self.hidden_layers,
                        self.learning_rate,
                        self.cluster_number,
                        self.cluster_labels], f)

    @classmethod
    def load(cls, filename):

        r"""

        This function can be used to load a saved MAF. For example

        .. code:: python

            from margarine.clustered import clusterMAF

            file = 'path/to/pickled/MAF.pkl'
            bij = clusterMAF.load(file)

        **Parameters:**

            filename: **string**
                | Path to the saved MAF.

        """

        with open(filename, 'rb') as f:
            data = pickle.load(f)
            theta, nn_weights, \
                sample_weights, \
                number_networks, \
                hidden_layers, \
                learning_rate, \
                cluster_number, \
                labels = data

        maf = cls(theta, weights=sample_weights,
                  number_networks=number_networks,
                  hidden_layers=hidden_layers,
                  learning_rate=learning_rate,
                  cluster_number=cluster_number,
                  cluster_labels=labels)
        maf(np.random.uniform(size=(1, theta.shape[-1])))
        for j in range(len(maf.flow)):
            for made, nnw in zip(maf.flow[j].mades, nn_weights[j]):
                made.set_weights(nnw)

        return maf



import numpy as np
from scipy.stats import gaussian_kde, norm
from margarine.processing import _forward_transform, _inverse_transform
from scipy.optimize import root_scalar
import tensorflow as tf
import pickle
import warnings
from tensorflow_probability import bijectors as tfb
import margarine
import anesthetic


class KDE(object):

    r"""
    This class is used to generate a KDE given a weighted set of samples,
    generate samples from that KDE, transform samples on the
    hypercube into samples on the KDE and save and load the KDE model.

    **Parameters:**

        theta: **numpy array or anesthetic.samples**
            | The samples from the probability distribution that we require the
                MAF to learn. This can either be a numpy array or an anesthetic
                NestedSamples or MCMCSamples object.

    **kwargs:**

        weights: **numpy array / default=np.ones(len(theta))**
            | The weights associated with the samples above. If an anesthetic
                NestedSamples or MCMCSamples object is passed the code
                draws the weights from this.

        bw_method: **str, scalar or callable**
            | The bandwidth for the KDE.

        theta_max: **numpy array**
            | The true upper limits of the priors used to generate the samples
                that we want the MAF to learn.

        theta_min: **numpy array**
            | As above but the true lower limits of the priors.

        parameters: **list of strings**
            | A list of the relevant parameters to train on. Only needed
                if theta is an anestehetic samples object. If not provided,
                all parameters will be used.
    **Attributes:**

    A list of some key attributes accessible to the user.

        kde: **Instance of scipy.stats.gaussian_kde**
            | Once the class has been initalised with a set of samples and
                their corresponding weights we can generate the kde using
                the following code

                .. code:: python

                    from bayesstats.kde import KDE
                    import numpy as np

                    theta = np.loadtxt('path/to/samples.txt')
                    weights = np.loadtxt('path/to/weights.txt')

                    KDE_class = KDE(theta, weights)
                    KDE_class.generate_kde()

                This is analogous to training a Normalising Flow (Bijector
                class). Once the KDE is generated it can be accessed via
                `KDE_class.kde`. Initialisation of the class and generation
                of the KDE are kept seperate to allow models to be saved and
                loaded effectively.

        theta_max: **numpy array**
            | The true upper limits of the priors used to generate the samples
                that we want the MAF to learn. If theta_max is not
                supplied as a kwarg, then this is is an approximate
                estimate.

        theta_min: **numpy array**
            | As above but for the true lower limits of the priors.
                If theta_max is not supplied as a kwarg, then this
                is an approximate estimate.

    """

    def __init__(self, theta, **kwargs):

        self.parameters = kwargs.pop('parameters', None)

        if isinstance(theta, 
                      (anesthetic.samples.NestedSamples, 
                       anesthetic.samples.MCMCSamples)):
            weights = theta.get_weights()
            if self.parameters:
                theta = theta[self.parameters].values
            else:
                if isinstance(theta, anesthetic.samples.NestedSamples):
                    self.parameters = theta.columns[:-3].values
                    theta = theta[theta.columns[:-3]].values
                else:
                    self.parameters = theta.columns[:-1].values
                    theta = theta[theta.columns[:-1]].values
        else:
            weights = kwargs.pop('weights', np.ones(len(theta)))

        self.theta = theta
        self.sample_weights = weights

        self.n = (np.sum(weights)**2)/(np.sum(weights**2))
        theta_max = np.max(theta, axis=0)
        theta_min = np.min(theta, axis=0)
        a = ((self.n-2)*theta_max-theta_min)/(self.n-3)
        b = ((self.n-2)*theta_min-theta_max)/(self.n-3)
        self.theta_min = kwargs.pop('theta_min', b)
        self.theta_max = kwargs.pop('theta_max', a)

        self.bw_method = kwargs.pop('bw_method', 'silverman')

    def generate_kde(self):

        r"""
        Function noramlises the input data into a standard normal parameter
        space and then generates a weighted KDE.
        """
        theta = tf.convert_to_tensor(self.theta, dtype=tf.float32)
        theta_min = tf.convert_to_tensor(self.theta_min, dtype=tf.float32)
        theta_max = tf.convert_to_tensor(self.theta_max, dtype=tf.float32)

        phi = _forward_transform(theta, theta_min, theta_max).numpy()
        mask = np.isfinite(phi).all(axis=-1)
        phi = phi[mask, :]
        weights_phi = self.sample_weights[mask]
        weights_phi /= weights_phi.sum()

        self.kde = gaussian_kde(
            phi.T, weights=self.sample_weights, bw_method=self.bw_method)

        return self.kde

    def __call__(self, u):

        r"""

        This function is used when calling the kde class to transform
        samples from the unit hypercube to samples on the kde.

        **Parameters:**

            u: **numpy array**
                | Samples on the uniform hypercube.

        """

        # generate useful parameters for __call__ function to transform
        # hypercube into samples on the KDE.
        S = self.kde.covariance
        mu = self.kde.dataset.T
        steps, s = [], []
        for i in range(mu.shape[-1]):
            step = S[i, :i] @ np.linalg.inv(S[:i, :i])
            steps.append(step)
            s.append((S[i, i] - step @ S[:i, i])**0.5)

        # transform samples from unit hypercube to kde
        transformed_samples = []
        for j in range(len(u)):
            x = u[j]
            y = np.zeros_like(x)
            for i in range(len(x)):
                m = mu[:, i] + steps[i] @ (y[:i] - mu[:, :i]).T
                y[i] = root_scalar(
                    lambda f:
                    (norm().cdf((f-m)/s[i])*self.kde.weights).sum()-x[i],
                    bracket=(mu[:, i].min()*2, mu[:, i].max()*2),
                    method='bisect').root
            transformed_samples.append(
                _inverse_transform(
                    tf.convert_to_tensor(y, dtype=tf.float32),
                    tf.convert_to_tensor(self.theta_min, dtype=tf.float32),
                    tf.convert_to_tensor(self.theta_max, dtype=tf.float32)))
        transformed_samples = np.array(transformed_samples)
        return transformed_samples

    def sample(self, length=1000):

        r"""

        Function can be used to generate samples from the KDE. It is much
        faster than the __call__ function but does not transform samples
        from the hypercube onto the KDE. It is however useful if we
        want to generate a large number of samples that can then be
        used to calulate the marginal statistics.

        **Kwargs:**

            length: **int / default=1000**
                | This should be an integer and is used to determine how many
                    samples are generated when calling the bijector.

        """
        x = self.kde.resample(length).T
        return _inverse_transform(
            tf.convert_to_tensor(x, dtype='float32'),
            tf.convert_to_tensor(self.theta_min, dtype=tf.float32),
            tf.convert_to_tensor(self.theta_max, dtype=tf.float32))

    def log_prob(self, params):

        """
        Function to caluclate the log-probability for a given KDE and
        set of parameters.

        While the density estimator has its own built in log probability
        function, a correction has to be applied for the transformation of
        variables that is used to improve accuracy when learning. The
        correction is implemented here.

        **Parameters:**

            params: **numpy array**
                | The set of samples for which to calculate the log
                    probability.

        """
        mins = self.theta_min.astype(np.float32)
        maxs = self.theta_max.astype(np.float32)

        transformed_x = _forward_transform(
            tf.convert_to_tensor(params, dtype=tf.float32),
            tf.convert_to_tensor(mins, dtype=tf.float32),
            tf.convert_to_tensor(maxs, dtype=tf.float32)).numpy()

        transform_chain = tfb.Chain([
            tfb.Invert(tfb.NormalCDF()),
            tfb.Scale(1/(maxs - mins)), tfb.Shift(-mins)])

        def norm_jac(y):
            return transform_chain.inverse_log_det_jacobian(
                y, event_ndims=0).numpy()

        correction = norm_jac(transformed_x)
        if params.ndim == 1:
            logprob = (self.kde.logpdf(transformed_x.T) -
                       np.sum(correction)).astype(np.float64)
        else:
            logprob = (self.kde.logpdf(transformed_x.T) -
                       np.sum(correction, axis=1)).astype(np.float64)

        return logprob

    def log_like(self, params, logevidence, prior=None, prior_weights=None):

        r"""
        This function should return the log-likelihood for a given set of
        parameters.

        It requires the logevidence from the original nested sampling run
        in order to do this and in the case that the prior is non-uniform
        prior samples should be provided.

        **Parameters:**

            params: **numpy array**
                | The set of samples for which to calculate the log
                    probability.

            logevidence: **float**
                | Should be the log-evidence from the full nested sampling
                    run with nuisance parameters.

            prior: **numpy array/default=None**
                | An array of prior samples corresponding to the prior. Default
                    assumption is that the prior is uniform which is
                    required if you want to combine likelihoods from different
                    experiments/data sets. In this case samples and prior
                    samples should be reweighted prior to any training.

        """

        if prior is None:
            warnings.warn('Assuming prior is uniform!')
            prior_logprob = np.log(np.prod(
                                   [1/(self.theta_max[i] - self.theta_min[i])
                                    for i in range(len(self.theta_min))]))
        else:
            self.prior = margarine.kde.KDE(prior, prior_weights)
            self.prior.generate_kde()
            prior_logprob_func = self.prior.log_prob
            prior_logprob = prior_logprob_func(params)

        posterior_logprob = self.log_prob(params)

        loglike = posterior_logprob + logevidence - prior_logprob

        return loglike

    def save(self, filename):

        r"""
        Function can be used to save an initalised version of the KDE class
        and it's assosiated generated KDE.

        **Parameters:**

            filename: **string**
                | Path in which to save the pickled KDE.
        """
        with open(filename, 'wb') as f:
            pickle.dump([self.theta, self.sample_weights, self.kde], f)

    @classmethod
    def load(cls, filename):

        r"""

        This function can be used to load a saved KDE. For example

        .. code:: python

            from margarine.kde import KDE

            file = 'path/to/pickled/bijector.pkl'
            KDE_class = KDE.load(file)

        **Parameters:**

            filename: **string**
                | Path to the saved KDE.

        """

        with open(filename, 'rb') as f:
            theta, sample_weights, kde = pickle.load(f)

        kde_class = cls(theta, weights=sample_weights)
        kde_class.kde = kde
        kde_class(np.random.uniform(0, 1, size=(2, theta.shape[-1])))

        return kde_class
